/*-
 * Copyright (c) 2005 Valentin Nechayev <netch@netch.kiev.ua>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main.c,v 1.18 2005/01/12 21:56:29 netch Exp $
 */

#include <sys/types.h>
#include <sys/disk.h>
#include <sys/endian.h>
#include <sys/uuid.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <inttypes.h>
#include <limits.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <uuid.h>
#include "sbuf.h"
#include "struc.h"
#include "assert.h"

sbuf sb_line;

/* Global: device block size. Equals to 512 in most cases. */
unsigned blocksize = 0;
/* Device size, in bytes. */
off_t devsize = 0;
/* Device size, in blocks. */
off_t devblocks = 0;

/* Device name */
char* g_devname = NULL;

int devfd = -1;

sbuf sb_gh_prim;
unsigned char *buf_gh_prim;
gpt_header gh_prim;
sbuf sb_array_prim;
int gpt_prim_valid = 0;       /* valid table in memory */
sbuf sb_gh_bkup;
unsigned char *buf_gh_bkup;
gpt_header gh_bkup;
sbuf sb_array_bkup;
int gpt_bkup_valid = 0;       /* valid table in memory */
off_t gpt_bkup_lba = 0;

sbuf sb_mbr;

off_t bkup_position = 0;       /* position from command line */
int f_doprim = 1;
int f_dobkup = 1;

/* Actual pointers for cmd_edit() */
gpt_header* ghp_actual;
sbuf* sb_array_actual;

typedef struct {
  const char* td_name;
  uuid_t td_uuid;
} id_name;

/* Some predefined partition types. Most data got from <sys/gpt.h>.
 * We should allow external data types file.
 */
id_name id_names[] = {
  { "EFI data", {0xc12a7328,0xf81f,0x11d2,0xba,0x4b,{0x00,0xa0,0xc9,0x3e,0xc9,0x3b}} },
  { "Protective MBR", {0x024dee41,0x33e7,0x11d3,0x9d,0x69,{0x00,0x08,0xc7,0x81,0xf3,0x9f}} },
  { "FreeBSD UFS", {0x516e7cb6,0x6ecf,0x11d6,0x8f,0xf8,{0x00,0x02,0x2d,0x09,0x71,0x2b}} },
  { "FreeBSD swap", {0x516e7cb5,0x6ecf,0x11d6,0x8f,0xf8,{0x00,0x02,0x2d,0x09,0x71,0x2b}} },
  { "FreeBSD label", {0x516e7cb4,0x6ecf,0x11d6,0x8f,0xf8,{0x00,0x02,0x2d,0x09,0x71,0x2b}} },
  { "FreeBSD vinum", {0x516e7cb8,0x6ecf,0x11d6,0x8f,0xf8,{0x00,0x02,0x2d,0x09,0x71,0x2b}} },
  { "Linux swap", {0x0657fd6d,0xa4ab,0x43c4,0x84,0xe5,{0x09,0x33,0xc8,0x4b,0x4f,0x4f}} },
  { "Linux LVM", {0xe6d6d379,0xf507,0x44c2,0xa2,0x3c,{0x23,0x8f,0x2a,0x3d,0xf9,0x28}} },
  { "Linux RAID", {0xa19d880f,0x05fc,0x4d3b,0xa0,0x06,{0x74,0x3f,0x0f,0x84,0x91,0x1e}} },
  { "Microsoft reserved", {0xe3c9e316,0x0b5c,0x4db8,0x81,0x7d,{0xf9,0x2d,0xf0,0x02,0x15,0xae}} },
  { "FreeBSD label", {0x516e7cb4,0x6ecf,0x11d6,0x8f,0xf8,{0x00,0x02,0x2d,0x09,0x71,0x2b}} },                                                                    
  { "FreeBSD vinum", {0x516e7cb8,0x6ecf,0x11d6,0x8f,0xf8,{0x00,0x02,0x2d,0x09,0x71,0x2b}} },                                                                    
  { "Linux swap", {0x0657fd6d,0xa4ab,0x43c4,0x84,0xe5,{0x09,0x33,0xc8,0x4b,0x4f,0x4f}} },                                                                       
  { "Linux LVM", {0xe6d6d379,0xf507,0x44c2,0xa2,0x3c,{0x23,0x8f,0x2a,0x3d,0xf9,0x28}} },                                                                        
  { "Linux RAID", {0xa19d880f,0x05fc,0x4d3b,0xa0,0x06,{0x74,0x3f,0x0f,0x84,0x91,0x1e}} },                                                                       
  { "Microsoft reserved", {0xe3c9e316,0x0b5c,0x4db8,0x81,0x7d,{0xf9,0x2d,0xf0,0x02,0x15,0xae}} },
  { "Microsoft basic data", {0xebd0a0a2,0xb9e5,0x4433,0x87,0xc0,{0x68,0xb6,0xb7,0x26,0x99,0xc7}} },
  { "Microsoft LDM metadata", {0x5808c8aa,0x7e8f,0x42e0,0x85,0xd2,{0xe1,0xe9,0x04,0x34,0xcf,0xb3}} },
  { "Microsoft LDM data", {0xaf9b60a0,0x1431,0x4f62,0xbc,0x68,{0x33,0x11,0x71,0x4a,0x69,0xad}} },
  /* Our own IDs */
  { "GPTE utility user type 1", {0x8cd8c4ad, 0x4cf1, 0x11d9, 0xbc, 0xf1, {0x00, 0x0e, 0x0c, 0x59, 0x51, 0xee}} },
  { "GPTE utility user type 2", {0x8cd8c4ae, 0x4cf1, 0x11d9, 0xbc, 0xf1, {0x00, 0x0e, 0x0c, 0x59, 0x51, 0xee}} },
  { "empty", {0, 0, 0, 0, 0, {0, 0, 0, 0, 0, 0}} },                             
  { NULL, {0, 0, 0, 0, 0, {0, 0, 0, 0, 0, 0}} }                                 
};                                                                              

static const unsigned char allzeros[16];

extern uint32_t crc32( const void*, size_t );

static int cmd_check( void );
static int cmd_create_empty( void );
static int cmd_edit( void );
static int cmd_erase( void );
static int cmd_fix( void );
static int cmd_print( void );
static int cmd_newids( void );
static const char* part_type_name( const uuid_t* );
static void iterate_partitions( gpt_header*, sbuf*,
    int (*)(void*, unsigned char*, unsigned), void*, int );
static int cmd_write_pmbr( void );

int
main( int argc, char *argv[] )
{
  char* command = NULL;
  int a;
  int nonoarg = 1; /* not after last possible argument */
  sbuf_zero(&sb_gh_prim);
  sbuf_zero(&sb_gh_bkup);
  sbuf_zero(&sb_array_prim);
  sbuf_zero(&sb_array_bkup);
  sbuf_zero(&sb_mbr);
  sbuf_zero(&sb_line);
  if( argc == 1 )
    errx( 1, "usage" );
  for( a = 1; a < argc; ++a ) {
    if( nonoarg ) {
      if( !strcmp( argv[a], "-backuppos" ) ) {
        char* rest;
        ++a; if( a >= argc ) errx( 1, "usage" );
        rest = NULL;
        bkup_position = strtoll( argv[a], &rest, 0 );
        if( rest && *rest ) errx( 1, "usage" );
        continue;
      }
      if( !strcmp( argv[a], "-blocksize" ) ) {
        char* rest;
        ++a; if( a >= argc ) errx( 1, "usage" );
        rest = NULL;
        blocksize = strtoll( argv[a], &rest, 0 );
        if( rest && *rest ) errx( 1, "usage" );
        if( blocksize < 512 || blocksize > 16384 ) /* voluntary */
          errx( 1, "invalid blocksize" );
        continue;
      }
      if( !strcmp( argv[a], "-primary" ) ) {
        f_dobkup = 0;
        continue;
      }
      if( !strcmp( argv[a], "-backup" ) ) {
        f_doprim = 0;
        continue;
      }
      if( !strcmp( argv[a], "--" ) ) {
        nonoarg = 0;
        continue;
      }
    }
    /* Here only if positional argument is possible. */
    if( !command ) {
      command = argv[a]; continue;
    }
    if( !g_devname ) {
      g_devname = argv[a]; continue;
    }
    errx( 1, "usage" );
  }
  if( !f_doprim && !f_dobkup )
    errx( 1, "usage: -primary and -backup are incompatible" );
  if( !command )
    errx( 1, "usage: no command" );
  if( !g_devname )
    errx( 1, "usage: no device" );
  /****/
  if( !strcmp( command, "check" ) )
    return cmd_check();
  if( !strcmp( command, "create_empty" ) )
    return cmd_create_empty();
  if( !strcmp( command, "edit" ) )
    return cmd_edit();
  if( !strcmp( command, "erase" ) )
    return cmd_erase();
  if( !strcmp( command, "newids" ) )
    return cmd_newids();
  if( !strcmp( command, "fix" ) )
    return cmd_fix();
  if( !strcmp( command, "print" ) )
    return cmd_print();
  if( !strcmp( command, "write_pmbr" ) )
    return cmd_write_pmbr();
  /****/
  /* No command found */
  errx( 1, "fatal: unknown command" );
}

void
open_device( int needwrite )
{
  if( !g_devname )
    errx( 1, "open_device: no devname" );
  devfd = open( g_devname, needwrite ? O_RDWR : O_RDONLY, 0 );
  if( devfd == -1 )
    err( 1, "open_device: open" );
  /* Get device characteristics */
#ifdef DIOCGSECTORSIZE
  if( !blocksize ) {
    u_int bs;
    if( ioctl( devfd, DIOCGSECTORSIZE, &bs ) == 0 )
      blocksize = bs;
  }
#endif
  if( !blocksize ) {
    warnx( "open_device: blocksize is unknown, falling back to 512" );
    blocksize = 512; /* fallback value */
  }
#ifdef DIOCGMEDIASIZE
  if( !devsize ) {
    off_t ds;
    if( ioctl( devfd, DIOCGMEDIASIZE, &ds ) == 0 ) {
      devsize = ds;
      devblocks = devsize / blocksize;
    }
  }
#endif
  /* Try seek? */
  if( !devsize ) {
    off_t seeksize = lseek( devfd, 0, SEEK_END );
      if( seeksize != -1 && seeksize > 0 ) {
        devsize = seeksize;
        devblocks = devsize / blocksize;
      }
  }
  if( !devsize && devblocks ) {
    devsize = devblocks * blocksize;
    if( devsize / blocksize != devblocks )
      errx( 1, "open_device: fatal: device size overflow" );
  }
  if( devsize && !devblocks ) {
    devblocks = devsize / blocksize;
  }
  if( !devsize )
    errx( 1, "open_device: fatal: device size is unknown" );
  if( devblocks < 100 )
    errx( 1, "fatal: device is too small to fit GPTs" );
  if( sbuf_extend( &sb_gh_prim, blocksize ) < 0 )
    err( 1, "sbuf_extend" );
  buf_gh_prim = sb_gh_prim.b_data;
  if( sbuf_extend( &sb_gh_bkup, blocksize ) < 0 )
    err( 1, "sbuf_extend" );
  buf_gh_bkup = sb_gh_bkup.b_data;
}

static int
read_tables( void )
{
  size_t list_prim_size;
  int bkup_kind;
  gpt_prim_valid = 0;
  gpt_bkup_valid = 0;
  /* Read LBA1 */
  if( !f_doprim ) goto read_gpt_bkup;
  if( lseek( devfd, blocksize, SEEK_SET ) == -1 )
    err( 1, "lseek" );
  if( read( devfd, buf_gh_prim, blocksize ) != (ssize_t)blocksize )
    err( 1, "read" );
  gpt_header_unpack( &gh_prim, buf_gh_prim );
  if( !gpt_header_check_sig( buf_gh_prim ) )
    goto read_gpt_bkup;
  if( !gpt_header_check_crc( buf_gh_prim ) )
    goto read_gpt_bkup;
  /* Read array */
  list_prim_size = gh_prim.gh_listcnt * gh_prim.gh_listelemsize;
  if( sbuf_extend( &sb_array_prim, list_prim_size ) < 0 )
    err( 1, "sbuf_extend" );
  if( lseek( devfd, blocksize*(off_t)gh_prim.gh_listlba, SEEK_SET ) == -1 )
    err( 1, "lseek(primary:partlist)" );
  if( read( devfd, sb_array_prim.b_data, list_prim_size )
      != (ssize_t)list_prim_size )
    err( 1, "read(primary:partlist)" );
  if( gpt_array_check_crc( sb_array_prim.b_data, list_prim_size,
      gh_prim.gh_listcrc ) )
    gpt_prim_valid = 1;
read_gpt_bkup:
  /* Generally, there are three places where to search for backup block:
   * 1. Place specified in command line
   * 2. Place specified as gh_altlba in primary table
   * 3. Last block of device
   * We will iterate all of them in order specified above.
   */
  for( bkup_kind = 1; bkup_kind <= 3; ++bkup_kind ) {
    off_t ap;
    size_t list_bkup_size;
    if( bkup_kind == 1 ) {
      ap = bkup_position;
      if( !ap ) continue;
      if( ap < 0 ) ap += devblocks;
    }
    else if( bkup_kind == 2 ) ap = gh_prim.gh_altlba;
    else ap = devblocks - 1;
    if( ap < 68 ) continue;
    if( lseek( devfd, blocksize * ap, SEEK_SET ) == -1 ) {
      warn( "lseek(backup:header)" ); continue;
    }
    if( read( devfd, buf_gh_bkup, blocksize ) != (ssize_t)blocksize ) {
      warn( "read(backup:header)" ); continue;
    }
    if( !gpt_header_check_sig( buf_gh_bkup ) ) continue;
    if( !gpt_header_check_crc( buf_gh_bkup ) ) continue;
    gpt_header_unpack( &gh_bkup, buf_gh_bkup );
    list_bkup_size = gh_bkup.gh_listcnt * gh_bkup.gh_listelemsize;
    if( sbuf_extend( &sb_array_bkup, list_bkup_size ) < 0 )
      err( 1, "sbuf_extend" );
    if( lseek( devfd, blocksize*(off_t)gh_bkup.gh_listlba, SEEK_SET ) == -1 ) {
      warn( "lseek(backup:partlist)" ); continue;
    }
    if( read( devfd, sb_array_bkup.b_data, list_bkup_size )
        != (ssize_t) list_bkup_size )
    {
      warn( "read(backup:partlist)" ); continue;
    }
    if( gpt_array_check_crc( sb_array_bkup.b_data, list_bkup_size,
        gh_bkup.gh_listcrc ) )
    {
      gpt_bkup_valid = 1;
      gpt_bkup_lba = ap;
      break;
    }
  }
  return 0; /* XXX */
}

static int
compare_tables( void )
{
  unsigned arraysize;
  /* Assume both tables are allowed to proceed, had been read and are valid */
  if( gh_prim.gh_revision != gh_bkup.gh_revision ||
      gh_prim.gh_hsize != gh_bkup.gh_hsize ||
      gh_prim.gh_mylba != gh_bkup.gh_altlba ||
      gh_prim.gh_altlba != gh_bkup.gh_mylba ||
      gh_prim.gh_fulba != gh_bkup.gh_fulba ||
      gh_prim.gh_lulba != gh_bkup.gh_lulba ||
      memcmp( &gh_prim.gh_guid, &gh_bkup.gh_guid, 16 ) ||
      gh_prim.gh_listcnt != gh_bkup.gh_listcnt ||
      gh_prim.gh_listelemsize != gh_bkup.gh_listelemsize )
    return 1;
  arraysize = gh_prim.gh_listcnt * gh_prim.gh_listelemsize;
  assert( sb_array_prim.b_size >= arraysize );
  assert( sb_array_bkup.b_size >= arraysize );
  return memcmp( sb_array_prim.b_data, sb_array_bkup.b_data, arraysize )
      ? 1 : 0;
}

static int
write_prim_table( void )
{
  size_t arraysize = gh_prim.gh_listcnt * gh_prim.gh_listelemsize;
  size_t arrayblocks = ( arraysize + blocksize - 1 ) / blocksize;
  off_t arraysize_b = (off_t)arrayblocks * blocksize;
  off_t arraylba;
  assert( sb_array_prim.b_data );
  assert( sb_array_prim.b_size >= arraysize );
  if( sbuf_extend( &sb_array_prim, arraysize_b ) < 0 )
    err( 1, "sbuf_extend" );
  arraylba = 2;
  assert( arraylba == (off_t)gh_prim.gh_listlba );
  if( lseek( devfd, arraylba * blocksize, SEEK_SET ) == -1 ) {
    warn( "lseek(primary:list)" ); return -1;
  }
  if( write( devfd, sb_array_prim.b_data, arraysize_b ) != arraysize_b ) {
    warn( "write(primary:list)" ); return -1;
  }
  gh_prim.gh_listcrc = crc32( sb_array_prim.b_data, arraysize );
  gpt_header_pack( &gh_prim, buf_gh_prim );
  if( lseek( devfd, blocksize, SEEK_SET ) == -1 ) {
    warn( "lseek(primary:header)" ); return -1;
  }
  if( write( devfd, buf_gh_prim, blocksize ) != (ssize_t)blocksize ) {
    warn( "write(primary:header)" ); return -1;
  }
  return 0;
}

static int
write_bkup_table( void )
{
  size_t arraysize = gh_bkup.gh_listcnt * gh_bkup.gh_listelemsize;
  size_t arrayblocks = ( arraysize + blocksize - 1 ) / blocksize;
  off_t arraysize_b = (off_t)arrayblocks * blocksize;
  off_t arraylba;
  assert( sb_array_bkup.b_data );
  assert( sb_array_bkup.b_size >= arraysize );
  if( sbuf_extend( &sb_array_bkup, arraysize_b ) < 0 )
    err( 1, "sbuf_extend" );
  assert( gpt_bkup_lba >= 4 ); /* for the infinite block size */
  arraylba = gpt_bkup_lba - (int)arrayblocks;
  assert( arraylba == (off_t)gh_bkup.gh_listlba );
  if( lseek( devfd, arraylba * blocksize, SEEK_SET ) == -1 ) {
    warn( "lseek(backup:list)" ); return -1;
  }
  if( write( devfd, sb_array_bkup.b_data, arraysize_b ) != arraysize_b ) {
    warn( "write(backup:list)" ); return -1;
  }
  gh_bkup.gh_listcrc = crc32( sb_array_bkup.b_data, arraysize );
  gpt_header_pack( &gh_bkup, buf_gh_bkup );
  if( lseek( devfd, gpt_bkup_lba * blocksize, SEEK_SET ) == -1 ) {
    warn( "lseek(backup:header)" ); return -1;
  }
  if( write( devfd, buf_gh_bkup, blocksize ) != (ssize_t)blocksize ) {
    warn( "write(backup:header)" ); return -1;
  }
  return 0;
}

static int
check_pmbr( void )
{
  unsigned char* mbr;
  int i;
  if( sbuf_extend( &sb_mbr, blocksize ) < 0 )
    err( 1, "sbuf_extend" );
  mbr = sb_mbr.b_data;
  if( lseek( devfd, 0, SEEK_SET ) == -1 ) {
    warn( "lseek(0)" ); return -1;
  }
  if( read( devfd, mbr, blocksize ) != (ssize_t)blocksize ) {
    warn( "read(mbr)" ); return -1;
  }
  if( mbr[510] != 0xAA || mbr[511] != 0x55 )
    return 0;
  for( i = 450; i < 512; i += 16 ) {
    if( mbr[i] != 0 && mbr[i] != 0xEE )
      return 0;
  }
  return 1;
}

static int
cmd_check( void )
{
  int r;
  open_device( 0 );
  if( read_tables() < 0 )
    errx( 1, "read_tables()" );
  if( f_doprim )
    printf( "Primary table is %svalid\n", gpt_prim_valid ? "" : "in" );
  if( f_dobkup ) {
    printf( "Backup table is %svalid (at: %jd,%jd)\n",
        gpt_bkup_valid ? "" : "in",
        (intmax_t) gpt_bkup_lba,
        gpt_bkup_lba ? (intmax_t) ( gpt_bkup_lba - devblocks ) : 0 );
  }
  if( f_doprim && f_dobkup && gpt_prim_valid && gpt_bkup_valid ) {
    if( compare_tables() )
      printf( "ATTN: tables are differ\n" );
  }
  r = check_pmbr();
  if( !r )
    printf( "ATTN: no PMBR found\n" );
  return 0;
}

static void
my_uuid_print( char* buf, const uuid_t* id )
{
  /* Buf must be at least 37 characters long, to fit uuid. */
  uint32_t urc;
  char* idstr;
  uuid_to_string( (uuid_t*)id, &idstr, &urc );
  if( urc == uuid_s_ok ) {
    strlcpy( buf, idstr, 37 );
    /* NB: DCE declares this should be rpc_string_free() */
    free( idstr );
  }
  else {
    snprintf( buf, 37, "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
        id->time_low, id->time_mid, id->time_hi_and_version, 
        id->clock_seq_hi_and_reserved, id->clock_seq_low,
        id->node[0], id->node[1], id->node[2],
        id->node[3], id->node[4], id->node[5] );
  }
}

static int
print_cb( void* cookie, unsigned char* rawge, unsigned np )
{
  /* XXX */
  gpt_elem ge;
  const char* part_type;
  char uup_type[40], uup_id[40];
  gpt_elem_unpack( &ge, rawge );
  printf( "Partition %3u:\n", np );
  printf( "  Position: %ju - %ju\n",
      (uintmax_t) ge.ge_begin, (uintmax_t) ge.ge_end );
  my_uuid_print( uup_type, &ge.ge_type );
  my_uuid_print( uup_id, &ge.ge_id );
  part_type = part_type_name( &ge.ge_type );
  if( part_type )
    printf( "  Type: %s (%s)\n", part_type, uup_type );
  else
    printf( "  Type: %s\n", uup_type );
  printf( "  Id:   %s\n", uup_id );
  printf( "  Attributes: %jx\n", (uintmax_t) ge.ge_attr );
  ++*(unsigned*)cookie;
  return 0;
}

static int
cmd_print( void )
{
  gpt_header* ghp;
  sbuf* arrayp;
  open_device( 0 );
  unsigned npart = 0;
  if( read_tables() < 0 )
    errx( 1, "read_tables()" );
  printf( "Primary table is %svalid\n", gpt_prim_valid ? "" : "in" );
  printf( "Backup table is %svalid (at: %jd,%jd)\n",
      gpt_bkup_valid ? "" : "in",
      (intmax_t) gpt_bkup_lba,
      gpt_bkup_lba ? (intmax_t) ( gpt_bkup_lba - devblocks ) : 0 );
  if( gpt_prim_valid ) {
    ghp = &gh_prim;
    arrayp = &sb_array_prim;
    printf( "Using primary table\n" );
  }
  else if( gpt_bkup_valid ) {
    ghp = &gh_bkup;
    arrayp = &sb_array_bkup;
    printf( "Using backup table because primary one is invalid\n" );
  }
  else
    errx( 1, "no valid table found, giving up" );
  printf( "Usable range: %ju-%ju\n",
     (uintmax_t) ghp->gh_fulba, (uintmax_t) ghp->gh_lulba );
  npart = 0;
  iterate_partitions( ghp, arrayp, print_cb, &npart, 0 );
  if( !npart )
    printf( "No partitions defined\n" );
  return 0;
}

static void
create_empty_primary_table( void )
{
  const int arraysize = 16384;
  int arraysize_a;
  unsigned listblocks;
  memset( buf_gh_prim, 0, blocksize );
  if( sbuf_extend( &sb_array_prim, arraysize ) < 0 )
    err( 1, "sbuf_extend" );
  memset( sb_array_prim.b_data, 0, arraysize );
  listblocks = ( blocksize - 1 + arraysize ) / blocksize;
  arraysize_a = listblocks * blocksize;
  memset( &gh_prim, 0, sizeof gh_prim );
  gh_prim.gh_revision = 0x00010000;
  gh_prim.gh_mylba = 1;
  gh_prim.gh_altlba = gpt_bkup_lba;
  gh_prim.gh_fulba = 2 + listblocks;
  gh_prim.gh_lulba = gpt_bkup_lba - listblocks - 1;
  if( gh_prim.gh_fulba >= gh_prim.gh_lulba )
    errx( 1, "cmd_create_empty(): fatal: too small disk" );
  gh_prim.gh_listlba = 2;
  gh_prim.gh_listcnt = 128;
  gh_prim.gh_listelemsize = 128;
  /* NB: this does not create disk GUID */
}

static void
create_empty_backup_table( void )
{
  const int arraysize = 16384;
  int arraysize_a;
  unsigned listblocks;
  memset( buf_gh_bkup, 0, blocksize );
  if( sbuf_extend( &sb_array_bkup, arraysize ) < 0 )
    err( 1, "sbuf_extend" );
  memset( sb_array_bkup.b_data, 0, arraysize );
  listblocks = ( blocksize - 1 + arraysize ) / blocksize;
  arraysize_a = listblocks * blocksize;
  memset( &gh_bkup, 0, sizeof gh_bkup );
  gh_bkup.gh_revision = 0x00010000;
  gh_bkup.gh_mylba = gpt_bkup_lba;
  gh_bkup.gh_altlba = 1;
  gh_bkup.gh_fulba = 2 + listblocks;
  gh_bkup.gh_lulba = gpt_bkup_lba - listblocks - 1;
  if( gh_bkup.gh_fulba >= gh_bkup.gh_lulba )
    errx( 1, "cmd_create_empty(): fatal: too small disk" );
  gh_bkup.gh_listlba = gpt_bkup_lba - listblocks;
  gh_bkup.gh_listcnt = 128;
  gh_bkup.gh_listelemsize = 128;
  /* NB: this does not create disk GUID */
}

static int
cmd_create_empty( void )
{
  open_device( 1 );
  /* Don't read current tables, we discard them... */
  if( bkup_position ) {
    gpt_bkup_lba = bkup_position;
    if( gpt_bkup_lba < 0 )
      gpt_bkup_lba += devblocks;
  }
  else
    gpt_bkup_lba = devblocks - 1;
  create_empty_primary_table();
  create_empty_backup_table();
  if( uuidgen( &gh_prim.gh_guid, 1 ) < 0 )
    err( 1, "uuidgen" );
  memmove( &gh_bkup.gh_guid, &gh_prim.gh_guid, sizeof gh_bkup.gh_guid );
  if( f_doprim && write_prim_table() < 0 )
    errx( 1, "write_prim_table()" );
  if( f_dobkup && write_bkup_table() < 0 )
    errx( 1, "write_bkup_table()" );
  return 0;
}

static int
cmd_erase( void )
{
  int bad = 0;
  open_device( 1 );
  if( read_tables() < 0 )
    err( 1, "read_tables()" );
  if( f_dobkup && gpt_bkup_valid ) {
    buf_gh_bkup[0] = '-'; /* erase signature */
    if( lseek( devfd, gpt_bkup_lba*blocksize, SEEK_SET ) == -1 ) {
      warn( "lseek(backup:header)" );
      bad = 1;
      goto erase_prim;
    }
    if( write( devfd, buf_gh_bkup, blocksize ) != (ssize_t)blocksize ) {
      warn( "write(backup:header)" );
      bad = 1;
    }
  }
erase_prim:
  if( f_doprim && gpt_prim_valid ) {
    buf_gh_prim[0] = '-'; /* erase signature */
    if( lseek( devfd, blocksize, SEEK_SET ) == -1 ) {
      warn( "lseek(primary:header)" );
      bad = 1;
      goto do_exit;
    }
    if( write( devfd, buf_gh_prim, blocksize ) != (ssize_t)blocksize ) {
      warn( "write(primary:header)" );
      bad = 1;
    }
  }
do_exit:
  return bad;
}

static void
fix_primary_table( void )
{
  if( sbuf_extend( &sb_array_prim, sb_array_bkup.b_size ) < 0 )
    err( 1, "sbuf_extend" );
  memmove( sb_array_prim.b_data, sb_array_bkup.b_data, sb_array_bkup.b_size );
  memmove( &gh_prim, &gh_bkup, sizeof gh_prim );
  gh_prim.gh_mylba = 1;
  gh_prim.gh_altlba = gpt_bkup_lba;
  gh_prim.gh_listlba = 2;
  memset( buf_gh_prim, 0, blocksize );
  if( write_prim_table() < 0 )
    errx( 1, "write_prim_table()" );
  gpt_prim_valid = 1;
}

static void
fix_backup_table( void )
{
  unsigned arraysize;
  /* We must select gpt_bkup_lba here. */
  if( bkup_position ) {
    gpt_bkup_lba = bkup_position;
    if( gpt_bkup_lba < 0 ) gpt_bkup_lba += devblocks;
  }
  else
      gpt_bkup_lba = devblocks - 1;
  if( sbuf_extend( &sb_array_bkup, sb_array_prim.b_size ) < 0 )
    err( 1, "sbuf_extend" );
  memmove( sb_array_bkup.b_data, sb_array_prim.b_data, sb_array_prim.b_size );
  memmove( &gh_bkup, &gh_prim, sizeof gh_bkup );
  gh_bkup.gh_mylba = gpt_bkup_lba;
  gh_bkup.gh_altlba = 1;
  arraysize = gh_bkup.gh_listcnt * gh_bkup.gh_listelemsize;
  gh_bkup.gh_listlba =
      gpt_bkup_lba - ( blocksize - 1 + arraysize ) / blocksize;
  memset( buf_gh_bkup, 0, blocksize );
  if( write_bkup_table() < 0 )
    errx( 1, "write_bkup_table()" );
  gpt_bkup_valid = 1;
}

static int
cmd_fix( void )
{
  if( !f_doprim || !f_dobkup )
    errx( 1, "fix: can fix only one table on another" );
  open_device( 1 );
  if( read_tables() < 0 )
    errx( 1, "read_tables()" );
  if( gpt_prim_valid && gpt_bkup_valid )
    errx( 0, "nothing to fix" );
  if( !gpt_prim_valid && !gpt_bkup_valid )
    errx( 1, "can't find correct table" );
  if( !gpt_prim_valid )
    fix_primary_table();
  if( !gpt_bkup_valid )
    fix_backup_table();
  return 0;
}

static int
yesno( void )
{
  int f = 1;
  int c, yes, no;
  do {
    yes = no = 0;
    fputs( f ? "?" : "(y/n)?", stdout ); fflush(stdout);
    do { c = getchar(); } while( c != EOF && c != '\n' && isspace(c) );
    if( c == EOF ) errx( 1, "EOF" );
    if( c == '\n' ) continue;
    if( c == 'y' || c == 'Y' ) yes = 1;
    if( c == 'n' || c == 'N' ) no = 1;
    do { c = getchar(); } while( c != EOF && c != '\n' );
    if( c == EOF ) errx( 1, "EOF" );
  } while( !yes && !no );
  return yes;
}

static int
mygetline( sbuf* sb, size_t maxlen )
{
  sb->b_pos = sb->b_len = 0;
  sb->b_rest = sb->b_size;
  for(;;) {
    int c = getchar();
    if( c == EOF ) return -1;
    if( !c ) c = ' ';
    if( c == '\n' ) c = 0;
    if( sb->b_rest == 0 ) {
      if( sbuf_extend( sb, sb->b_size * 2 + 20 ) < 0 )
        err( 1, "sbuf_extend" );
      sb->b_rest = sb->b_size - sb->b_len;
      assert( sb->b_rest > 0 );
    }
    sb->b_data[sb->b_len] = c;
    ++sb->b_len;
    --sb->b_rest;
    if( sb->b_len > maxlen ) { errno = E2BIG; return -1; }
    if( !c ) break;
  }
  return 0;
}

static void
ce_print_header( void )
{
  /****/
}

static void
iterate_partitions( gpt_header* gh, sbuf* sb_array,
    int (*cb)(void*, unsigned char*, unsigned), void* cookie,
    int b_with_empty )
{
  unsigned npz;
  for( npz = 0; ; ++npz ) {
    unsigned ofcurr, ofnext;
    unsigned char* p;
    ofcurr = npz * gh->gh_listelemsize;
    ofnext = ofcurr + gh->gh_listelemsize;
    if( ofcurr >= sb_array->b_size || ofnext > sb_array->b_size )
      break;
    p = sb_array->b_data + ofcurr;
    if( !b_with_empty && !memcmp( p, allzeros, 16 ) )
      continue;
    if( (*cb)(cookie,p,npz+1) )
      break;
  }
}

static const char*
part_type_name( const uuid_t* u )
{
  int i = 0;
  for( i = 0; id_names[i].td_name; ++i ) {
    if( !memcmp( u, &id_names[i].td_uuid, sizeof( *u ) ) )
      return id_names[i].td_name;
  }
  return NULL;
}

typedef struct ce_print_cb_data {
  int nfound;
  int f_long;
} ce_print_cb_data;

static int
ce_print_cb( void* cookie, unsigned char* p, unsigned np )
{
  const char* typename;
  gpt_elem elem;
  char uup_type[40], uup_id[40];
  ce_print_cb_data* cb_data = (ce_print_cb_data*) cookie;
  gpt_elem_unpack( &elem, p );
  my_uuid_print( uup_type, &elem.ge_type );
  typename = part_type_name( &elem.ge_type );
  if( cb_data->f_long ) {
    printf( "Partition %3u:\n", np );
    printf( "  Type: %s (%s)\n", typename ? typename : "unknown", uup_type );
    my_uuid_print( uup_id, &elem.ge_id );
    printf( "  Id:   %s\n", uup_id );
    printf( "  Begin: %ju\n", (uintmax_t) elem.ge_begin );
    printf( "  End: %ju\n", (uintmax_t) elem.ge_end );
    printf( "  Attributes: %jx\n", (uintmax_t) elem.ge_attr );
    /* XXX add more info */
  }
  else {
    printf( "%3u: %ju-%ju: %s (%s)\n",
        np, (uintmax_t) elem.ge_begin, (uintmax_t) elem.ge_end,
        typename ? typename : "unknown",
        uup_type );
  }
  ++cb_data->nfound;
  return 0;
}

static void
ce_print( int f_long )
{
  ce_print_cb_data cb_data;
  cb_data.nfound = 0;
  cb_data.f_long = f_long;
  printf( "Device %s", g_devname );
  printf( " is %ju blocks long", (uintmax_t) devblocks );
  printf( ", %ju bytes", (uintmax_t) devsize );
  if( devsize >= (off_t) 10000000000ull ) {
    printf( " (%.3f GB, %.3f GiB)",
        devsize / 1.0e9,
        devsize / 10731741824.0 );
  }
  else if( devsize >= 10000000 ) {
    printf( " (%.3f MB, %.3f MiB)",
        devsize / 1.0e6,
        devsize / 1048576.0 );
  }
  printf( "\n" );
  printf( "Blocksize is %u\n", (unsigned) blocksize );
  iterate_partitions( ghp_actual, sb_array_actual,
      ce_print_cb, &cb_data, 0 );
  if( !cb_data.nfound )
    printf( "Table is empty.\n" );
}

static int
read_part_type( uuid_t* t, const char* prompt )
{
  uint32_t urc;
  int i;
r:
  printf( "%s (l - list known, r - new random, <empty> - cancel)?", prompt );
  fflush( stdout );
  if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
  if( !sb_line.b_data[0] )
    return 0;
  if( !strcmp( sb_line.b_data, "r" ) ) {
    if( uuidgen( t, 1 ) < 0 ) {
      warn( "uuidgen" ); goto r;
    }
    return 1;
  }
  if( 0 != strcmp( sb_line.b_data, "l" ) ) {
    uuid_from_string( sb_line.b_data, t, &urc );
    if( urc != uuid_s_ok )
      goto r;
    return 1;
  }
  /* Print list and ask number */
  for( i = 0; id_names[i].td_name; ++i ) {
    char* ustr;
    uuid_to_string( &id_names[i].td_uuid, &ustr, &urc );
    printf( "%02d: %s (%s)\n", i+1, id_names[i].td_name, ustr );
  }
  printf( "?" ); fflush( stdout );
  if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
  i = strtoul( sb_line.b_data, NULL, 0 );
  if( i == 0 ) goto r;
  /* XXX check for array size */
  --i;
  *t = id_names[i].td_uuid;
  return 1;
}

static void
ce_new( void )
{
  int rc, pn;
  char *p2;
  char* ppos;
  gpt_elem ge;
  for(;;) {
    printf( "Partition number (0 - return)?" ); fflush( stdout );
    if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
    pn = strtol( sb_line.b_data, &p2, 0 );
    if( p2 && *p2 ) { printf( "Syntax error\n" ); continue; }
    if( pn < 0 ) { printf( "Invalid value\n" ); continue; }
    break;
  }
  if( pn == 0 ) return;
  if( pn > 128 ) {
    /* XXX extern array if needed */
    printf( "Partition number is too big\n" ); return;
  }
  ppos = sb_array_actual->b_data +
      ( pn - 1 ) * ghp_actual->gh_listelemsize;
  if( memcmp( ppos, allzeros, 16 ) != 0 )
  {
    printf( "Error: partition exists\n" ); return;
  }
  memset( &ge, 0, sizeof ge );
  /* XXX check allowed range for begin and end */
  printf( "Starting LBA?" ); fflush( stdout );
  if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
  ge.ge_begin = strtoumax( sb_line.b_data, &p2, 0 );
  if( p2 && *p2 ) { printf( "Syntax error\n" ); return; }
  printf( "Ending LBA?" ); fflush( stdout );
  if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
  ge.ge_end = strtoumax( sb_line.b_data, &p2, 0 );
  if( p2 && *p2 ) { printf( "Syntax error\n" ); return; }
  rc = read_part_type( &ge.ge_type, "Partition type" );
  if( rc <= 0 ) return;
  if( uuidgen( &ge.ge_id, 1 ) < 0 ) err( 1, "uuidgen" );
  gpt_elem_pack( &ge, ppos );
  printf( "Created.\n" );
}

static int
write_and_quit( void )
{
  /* Write table and quit. */
  if( f_doprim ) {
    if( write_prim_table() < 0 )
      errx( 1, "write_prim_table()" );
  }
  if( f_doprim && f_dobkup )
    fix_backup_table();
  if( f_dobkup ) {
    if( write_bkup_table() < 0 )
      errx( 1, "write_prim_table()" );
  }
  return 0;
}

static unsigned char*
ceaux_get_part_ptr( int b_with_empty )
{
  char* p2;
  int pn;
  unsigned ofs;
  unsigned char* ptr;
  printf( "Partition number (0 - return)?" ); fflush( stdout );
  if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
  pn = strtol( sb_line.b_data, &p2, 0 );
  if( p2 && *p2 ) { printf( "Syntax error\n" ); return NULL; }
  if( pn < 0 ) { printf( "Invalid value\n" ); return NULL; }
  if( pn == 0 ) return NULL;
  if( (unsigned)pn > ghp_actual->gh_listcnt ) {
    printf( "Partition number is too big\n" ); return NULL;
  }
  ofs = ghp_actual->gh_listelemsize * pn;
  assert( ofs <= sb_array_actual->b_size );
  ofs -= ghp_actual->gh_listelemsize;
  ptr = sb_array_actual->b_data + ofs;
  if( !b_with_empty && !memcmp( ptr, allzeros, 16 ) )
  {
    printf( "No such partition\n" ); return NULL;
  }
  return ptr;
}

static void
ce_delete( void )
{
  unsigned char* ptr = ceaux_get_part_ptr( 0 );
  if( !ptr ) return;
  memset( ptr, 0, ghp_actual->gh_listelemsize );
}

static void
ce_settype( void )
{
  gpt_elem ge;
  unsigned char* ptr = ceaux_get_part_ptr( 0 );
  if( !ptr ) return;
  gpt_elem_unpack( &ge, ptr );
  if( read_part_type( &ge.ge_type, "New type" ) < 0 ) return;
  gpt_elem_pack( &ge, ptr );
}

static void
ce_setid( void )
{
  gpt_elem ge;
  uuid_t newid;
  uint32_t urc;
  unsigned char* ptr = ceaux_get_part_ptr( 0 );
  if( !ptr ) return;
  gpt_elem_unpack( &ge, ptr );
  printf( "New id?" ); fflush( stdout );
  if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
  uuid_from_string( sb_line.b_data, &newid, &urc );
  if( urc == uuid_s_ok ) {
    memmove( &ge.ge_id, &newid, sizeof newid );
    gpt_elem_pack( &ge, ptr );
    return;
  }
  printf( "Error: invalid uuid\n" );
}

static void
ce_setattrs( void )
{
  gpt_elem ge;
  uuid_t newid;
  char* p2;
  unsigned char* ptr = ceaux_get_part_ptr( 0 );
  if( !ptr ) return;
  gpt_elem_unpack( &ge, ptr );
  printf( "New attributes?" ); fflush( stdout );
  if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
  ge.ge_attr = strtoumax( sb_line.b_data, &p2, 0 );
  if( p2 || *p2 ) {
    printf( "Error: syntax\n" );
    return;
  }
  memmove( &ge.ge_id, &newid, sizeof newid );
  gpt_elem_pack( &ge, ptr );
}

static int
cmd_edit( void )
{
  char lbuf[100];
  if( !isatty(0) || !isatty(1) )
    errx( 1, "usage: edit is interactive and requires terminal" );
  open_device( 1 );
  if( read_tables() < 0 )
    err( 1, "read_tables()" );
  if( f_doprim && f_dobkup && !gpt_prim_valid && gpt_bkup_valid ) {
    printf( "Primary table is invalid. It should be fixed before any editing,\n"
            "with data from backup table. Proceed to fix?\n" );
    fflush(stdout);
    if( !yesno() ) {
      printf( "Can't fix primary table, giving up\n" );
      return 1;
    }
    fix_primary_table();
  }
  if( f_doprim && f_dobkup && gpt_prim_valid && !gpt_bkup_valid ) {
    printf(
        "Backup table is invalid. It should be fixed before any editing,\n"
        "with data from primary table. Proceed to fix?\n" );
    fflush(stdout);
    if( !yesno() ) {
      printf( "Can't fix backup table, giving up\n" );
      return 1;
    }
    fix_backup_table();
  }
  if( !gpt_prim_valid && !gpt_bkup_valid ) {
    printf( "No tables found.\nCreate?" ); fflush(stdout);
    if( !fgets( lbuf, sizeof lbuf, stdin ) ) err( 1, "fgets" );
    if( lbuf[0] != 'Y' && lbuf[0] != 'y' ) {
      printf( "Giving up.\n" ); return 1;
    }
    create_empty_primary_table();
    create_empty_backup_table();
    if( uuidgen( &gh_prim.gh_guid, 1 ) < 0 )
      err( 1, "uuidgen" );
    memmove( &gh_bkup.gh_guid, &gh_prim.gh_guid, sizeof gh_bkup.gh_guid );
    /****/
    /****/
  }
  if( f_doprim && f_dobkup &&
      gpt_prim_valid && gpt_bkup_valid && compare_tables() )
  {
    printf(
        "Backup table is valid but differs from main one. It should be fixed\n"
        "before any editing, with data from primary table. Proceed to fix?\n" );
    fflush(stdout);
    if( !yesno() ) {
      printf( "Can't fix backup table, giving up\n" );
      return 1;
    }
    fix_backup_table();
  }
  /* Tables are identical. Start editing */
  ghp_actual = f_doprim ? &gh_prim : &gh_bkup;
  sb_array_actual = f_doprim ? &sb_array_prim : &sb_array_bkup;
  for(;;) {
    char* line;
    printf( "Command (h for help):" ); fflush(stdout);
    if( mygetline( &sb_line, 100 ) < 0 ) err( 1, "mygetline" );
    line = sb_line.b_data;
    if( !strcmp( line, "h" ) ) {
      printf( "n   - new partition\n" );
      printf( "d   - delete partition\n" );
      printf( "p   - print table\n" );
      printf( "pl  - print table in long format\n" );
      printf( "ph  - print header\n" );
      printf( "sa  - set partition attributes value\n" );
      printf( "si  - set partition ID\n" );
      printf( "st  - set partition type\n" );
      printf( "w   - save table and quit\n" );
      printf( "q   - quit\n" );
      continue;
    }
    if( !strcmp( line, "n" ) ) { ce_new(); continue; }
    if( !strcmp( line, "d" ) ) { ce_delete(); continue; }
    if( !strcmp( line, "p" ) ) { ce_print(0); continue; }
    if( !strcmp( line, "pl" ) ) { ce_print(1); continue; }
    if( !strcmp( line, "ph" ) ) { ce_print_header(); continue; }
    if( !strcmp( line, "sa" ) ) { ce_setattrs(); continue; }
    if( !strcmp( line, "si" ) ) { ce_setid(); continue; }
    if( !strcmp( line, "st" ) ) { ce_settype(); continue; }
    if( !strcmp( line, "w" ) ) { return write_and_quit(); }
    if( !strcmp( line, "q" ) ) return 0;
    printf( "Unknown command\n" );
  }
}

static int
newids_cb( void* cookie, unsigned char* ptr, unsigned np )
{
  gpt_elem ge;
  gpt_elem_unpack( &ge, ptr );
  /*ARGSUSED*/
  (void) cookie; (void) np;
  if( uuidgen( &ge.ge_id, 1 ) < 0 )
    err( 1, "uuidgen" );
  gpt_elem_pack( &ge, ptr );
  return 0;
}

static int
cmd_newids( void )
{
  open_device( 1 );
  if( read_tables() < 0 )
    err( 1, "read_tables()" );
  if( f_doprim && f_dobkup && compare_tables() )
    errx( 1, "first fix tables because they are differ" );
  ghp_actual = f_doprim ? &gh_prim : &gh_bkup;
  sb_array_actual = f_doprim ? &sb_array_prim : &sb_array_bkup;
  if( uuidgen( &ghp_actual->gh_guid, 1 ) < 0 )
    err( 1, "uuidgen" );
  iterate_partitions( ghp_actual, sb_array_actual,
      &newids_cb, NULL, 0 );
  return write_and_quit();
}

static int
cmd_write_pmbr( void )
{
  open_device( 1 );
  unsigned char* mbr;
  if( sbuf_extend( &sb_mbr, blocksize ) < 0 )
    err( 1, "sbuf_extend" );
  mbr = sb_mbr.b_data;
  if( lseek( devfd, 0, SEEK_SET ) == -1 ) {
    warn( "lseek(0)" ); return -1;
  }
  if( read( devfd, mbr, blocksize ) != (ssize_t)blocksize ) {
    warn( "read(mbr)" ); return -1;
  }
  mbr[blocksize-2] = 0xAA;
  mbr[blocksize-1] = 0x55;
  mbr[466] = 0;
  mbr[482] = 0;
  mbr[498] = 0;
  mbr[450] = 0xEE;
  /* XXX fill size info? */
  if( lseek( devfd, 0, SEEK_SET ) == -1 ) {
    warn( "lseek(0)" ); return -1;
  }
  if( write( devfd, mbr, blocksize ) != (ssize_t)blocksize ) {
    warn( "write(mbr)" ); return -1;
  }
  return 0;
}
