/*-
 * Copyright (c) 2005 Valentin Nechayev <netch@netch.kiev.ua>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: struc.h,v 1.3 2005/01/12 21:57:50 netch Exp $
 */

#ifndef struc_H_defined
#define struc_H_defined

#include <sys/types.h>
#include <sys/uuid.h>
/* Unpacked structures. */

/* GPT header */
typedef struct gpt_header {
  char gh_sig[8];             /* 0x00 Signature */
  uint32_t gh_revision;       /* 0x08 Revision */
  uint32_t gh_hsize;          /* 0x0C HeaderSize */
  uint32_t gh_hcrc;           /* 0x10 HeaderCRC32 */
  uint32_t gh_reserved;
  uint64_t gh_mylba;          /* 0x18 MyLBA */
  uint64_t gh_altlba;         /* 0x20 AlternateLBA */
  uint64_t gh_fulba;          /* 0x28 FirstUsableLBA */
  uint64_t gh_lulba;          /* 0x30 LastUsableLBA */
  uuid_t   gh_guid;           /* 0x38 DiskGUID */
  uint64_t gh_listlba;        /* 0x48 PartitionEntryLBA */
  uint32_t gh_listcnt;        /* 0x50 NumberOfPartitionEntries */
  uint32_t gh_listelemsize;   /* 0x54 SizeOfPartitionEntry */
  uint32_t gh_listcrc;        /* 0x58 PartitionEntryArrayCRC32 */
} gpt_header;

void gpt_header_unpack( gpt_header*, const unsigned char* );
void gpt_header_pack( const gpt_header*, unsigned char* );
int gpt_header_check_sig( unsigned char* );
int gpt_header_check_crc( unsigned char* );

/* GPT partition table element */
typedef struct gpt_elem {
  uuid_t   ge_type;           /* PartitionTypeGUID */
  uuid_t   ge_id;             /* Unique Partition GUID */
  uint64_t ge_begin;
  uint64_t ge_end;
  uint64_t ge_attr;
  char     ge_name[72];
  char     ge_extra[1];       /* Keeping extra data, anchor is here */
} gpt_elem;

void gpt_elem_unpack( gpt_elem*, const unsigned char* );
void gpt_elem_pack( const gpt_elem*, unsigned char* );

int gpt_array_check_crc( unsigned char*, size_t, uint32_t );

#endif
